# README

## Manuscript

**Title**: MULTIMODAL MACHINE LEARNING FOR MODELING INFANT HEAD CIRCUMFERENCE, MOTHERS’ MILK COMPOSITION, AND THEIR SHARED ENVIRONMENT

**Authors**: Martin Becker, Kelsey Fehr, Stephanie Goguen, Kozeta Miliku, Catherine Field, Bianca Roberston, Chloe Yonemitsu, Lars Bode, Elinor Simons, Jean Marshall, Bassel Dawod, Piushkumar Mandhane, Stuart E. Turvey, Theo J. Moraes, Padmaja Subbarao, Natalie Rodriguez, Nima Aghaeepour, Meghan B. Azad


**Summary**: Links between human milk (HM) and infant development are poorly understood and often focus on individual HM components. Here we apply multi-modal predictive machine learning to study HM and head circumference (a proxy for brain development) among 1022 mother-infant dyads of the CHILD Cohort. We integrated HM data (19 oligosaccharides, 28 fatty acids, 3 hormones, 28 chemokines) with maternal and infant demographic, health, dietary, and home environment data. Head circumference was significantly predictable at 3 and 12 months. Two especially associated features were HM n3-polyunsaturated fatty acid C22:6n3 (docosahexaenoic acid, DHA; p=9.6e-05) and maternal intake of fish (p=4.1e-03), an essential dietary source of DHA with established relationships to brain function. Thus, using a systems biology approach, we identified meaningful relationships between HM and brain development. This validates our statistical approach, gives credence to the novel associations we observed, and sets the foundation for further research with additional cohorts and HM analytes.

## Data

For data access please contact _Meghan B. Azad_.

Place the following files from the provided data archive in the `data/processed` folder:

* `data_milk.csv`
* `codebook.csv`

## Code

Use the following commands the reproduce the results from the manuscript. You will find the results in the `_out` folder. Alternatively, the processed notebooks can be found in the `notebooks/publication-2023` folder.

```bash
# setup the Python environment
conda env create -f environment.yml  # you can also use mamba for faster execution
conda activate nalab-milk
python -m ipykernel install --user --name nalab-milk 
pip install -e .

# run the code
jupyter nbconvert --to notebook --execute notebooks/publication-2023/1___data-analysis.ipynb
jupyter nbconvert --to notebook --execute notebooks/publication-2023/2___predict-train
jupyter nbconvert --to notebook --execute notebooks/publication-2023/3___predict-analyze.ipynb
```

## Note

This project has been set up using PyScaffold 3.2.1. For details and usage
information on PyScaffold see https://pyscaffold.org/.
