import warnings
import pandas as pd


def to_excel(df, filename, apply_style=None, **to_excel_kwargs):
    """
    source: https://stackoverflow.com/a/40535454
    """
    
    if apply_style is not None:
        df_styled = apply_style(df)
    else:
        df_styled = df
    
    try:
        # write to excel
        writer = pd.ExcelWriter(filename, engine='xlsxwriter')
        df_styled.to_excel(writer, **to_excel_kwargs)

        # set columns widths
        worksheet = writer.sheets["Sheet1"] # pull worksheet object
        for idx, col in enumerate(df):  # loop through all columns
            series = df[col]
            max_len = max((
                series.astype(str).map(len).max(),  # len of largest item
                len(str(series.name))  # len of column name/header
                )) + 1  # adding a little extra space
            worksheet.set_column(idx, idx, max_len)  # set column width

        writer.save()
    except:
        warnings.warn("ERROR: `xslxwriter` missing. Excel not written.")
    finally:
        return df_styled
    
