"""
This is intended to analyze, collect and visualize 
information about features in a dataset and 
their relation to a target.
But I think it's a bit of an overkill 
as it is easier and more clear 
to just calculate these statistics directly 
in the corresponding script.
"""

import collections
from .stats.univariate import calculate_pairwise_comparisons
from .visualization.network import nx_plot

import numpy as np

import scipy as sp
import scipy.stats

import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

import sklearn
import sklearn.cluster
import sklearn.discriminant_analysis
import sklearn.impute

import matplotlib.image

from matplotlib import gridspec
import matplotlib

from .utils import dicts
from .cor import calculate_correlation_parallel

# def binary_difference(
#         binary_labels,
#         data,
#         fisher_exact=False,
#         chi2=False,
#         ranksums=True, 
#         mannwhitneyu=False, 
#         custom_difference_function=None, 
#         p1_on_error=False,
#         multipletests="bonferroni",
#         exclude_labels=True,
#         return_dataframe=False,
#         samples_fraction=None,
#         samples_n=100,
#         overall_stats_label="__overall__"):
#     """Given a binary label for each row,
#     this function tests if the values in each each column grouped by these labels
#     come from the same distribution. 
#     Various test are supported including Pearson's Chi2, Ranksums (Wilcoxon), and Mann-Whitney-U.   
    
#     Parameters
#     ----------
#     binary_labels : str or np.array or pd.Series
#         binary labels
#     data : pd.DataFrame
#         the columns to compare (may include binary labels)
#     fisher_exact : bool or dict, optional
#         whether to calculate Fisher's exact test; if a dict is given,
#         it is passed as parameters to `scipy.stats.chi2_contingency`;
#         note that this test is only run if the columns values are binary (if not it is set to None)
#         (the default is False, which does not run the test)
#     chi2 : bool or dict, optional
#         whether to calculate Person's Chi2; if a dict is given,
#         it is used parameters of `scipy.stats.chi2_contingency`
#         (the default is False, which does not run the test)
#     ranksums : bool, optional
#         whether to calculate Wilcoxon Ranksums; if a dict is given,
#         it is used parameters of `scipy.stats.ranksums`
#         (the default is True, which runs the test)
#     mannwhitneyu : bool, optional
#         whether to calculate Mann-Whitney U; if a dict is given,
#         it is used parameters of `scipy.stats.mannwhitneyu`
#         (the default is False, which does not run the test)
#     custom_difference_function : difference function or list of name-difference function tuples ([("custom1", diff_function1), ...]), optional
#         difference_function(
#             pd.DataFrame({ "bin": [...], "values": [...]  }),
#             sorted([binary label 0, binary label 1]),
#             pd.DataFrame with label 0,
#             pd.DataFrame with label 1)
#                 return  { "statistic": chi2[0], "pvalue": chi2[1], "data": chi2 }
#         (the default is None, which results in no additional tests)
#     p1_on_error : bool, optional
#         If set to `True` and an error is raised,
#         we set `{ "statistic": None, "pvalue": 1, "error": True, "data": <exception> }`
#         (the default is False, which raises each exception/warning)
#     multipletests : str or list of str
#         Multiple comparison tests for pvalues of statistical tests
#         (cf. https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html)
#     exclude_labels : bool, optional
#         If the labels are part of the data and this is set to `False`,
#         we include the label in the results
#         (the default is True, which excludes the labels)
#     return_dataframe: bool, optional
#         If set to true, the method returns a simplified version of this as a pandas.DataFrame
    
#     Returns
#     -------
#     dict
#         {
#             "column_name": {
#                 "stats": {
#                     "<value1>": {
#                         "std": 123,
#                         "mean": 123,
#                         "median": 123,
#                         "n": 123,
#                     },
#                     "<value2>": { ... }
#                 },
#                 "tests": {
#                     "test": {
#                         "statistic": 123,
#                         "pvalue": 0.05,
#                         "data": <complete test result>,
#                         ["error": True] # optional; only set if True
#                     }
#                 }
#             }
#         }
#     """

#     if isinstance(binary_labels, str):
#         if data is None:
#             raise Exception("If 'binary_labels' is a String, 'data' must be set.")
#         binary_data = data[binary_labels]
#     else:
#         binary_data = binary_labels
    
#     differences = {}

#     for c in data.columns:

#         # skip binary labels column
#         if exclude_labels and isinstance(binary_labels, str) and c == binary_labels:
#             continue

#         # check binary
#         labels = sorted(binary_data.unique())
#         if len(labels) != 2:
#             raise Exception("Labels must be binary. However we detected {} distinct values: {}". format(len(labels), labels))

#         # prepare data
#         df = pd.DataFrame({"bin": binary_data, "values": data[c]})
#         c0 = df[df["bin"] == labels[0]]
#         c1 = df[df["bin"] == labels[1]]

#         # init differences
#         def stats(c):
#             return [
#                 ("median", np.nanmedian(c)),
#                 ("mean", np.nanmean(c)),
#                 ("std", np.nanstd(c)),
#                 ("n", (~np.isnan(c)).sum())
#             ]
#         differences[c] = {
#             "stats": {
#                 overall_stats_label: collections.OrderedDict([*stats(data[c]), *[
#                     ("n_na", data[c].isna().sum()),
#                     ("n_unique", len(data[c][~np.isnan(data[c])].unique())),
#                     ("binary", len(data[c][~np.isnan(data[c])].unique()) == 2),
#                     ("unary", len(data[c][~np.isnan(data[c])].unique()) == 1)]]),
#                 labels[0]: collections.OrderedDict(stats(c0["values"])),
#                 labels[1]: collections.OrderedDict(stats(c1["values"]))
#             },
#             "tests": {}
#         }

#         # tests
#         difference_map = {}
#         differences[c]["tests"] = difference_map

#         def calc_diff_generic(difference_function, df, labels, c0, c1):
#             try:
#                 return difference_function(df, labels, c0, c1)
#             except Exception as e:
#                 if p1_on_error:
#                     return {"statistic": None, "pvalue": 1, "error": True, "data": e}
#                 else:
#                     raise e

#         def sample_diff(name, difference_function):

#             split = sklearn.model_selection.StratifiedShuffleSplit(
#                 n_splits=samples_n, train_size=samples_fraction, test_size=1-samples_fraction)
#             samples = [s for s, _ in split.split(df["values"], df["bin"])]

#             def single_sample(idx):
#                 df_sample = df.iloc[idx, :]
#                 c0_sample = df_sample[df_sample["bin"] == labels[0]]
#                 c1_sample = df_sample[df_sample["bin"] == labels[1]]
#                 return calc_diff_generic(difference_function, df_sample, labels, c0_sample, c1_sample)

#             results = [single_sample(idx) for idx in samples]

#             return {
#                 "statistic": np.mean([ s["statistic"] for s in results]),
#                 "statistic_std": np.std([ s["statistic"] for s in results]),
#                 "pvalue": np.mean([ s["pvalue"] for s in results]),
#                 "pvalue_std": np.std([ s["pvalue"] for s in results]),
#                 "data": {
#                     "samples": results,
#                     "splits": samples
#                 }
#             }

#         def calc_diff(name, difference_function):

#             difference_map[name] = calc_diff_generic(difference_function, df, labels, c0, c1)

#             if samples_fraction is not None:
#                 difference_map[name]["___sampled"] = sample_diff(name, difference_function)

#         # pearson chi2; only works on binary data values
#         if fisher_exact:

#             # set parameters
#             fisher_parameters = {}
#             if isinstance(fisher_exact, dict):
#                 fisher_parameters = fisher_exact

#             # define function
#             def fisher_f(df, labels, c0, c1):
#                 table = df.reset_index().groupby(["bin", "values"]).count().unstack().fillna(0)

#                 if table.shape[1] == 2:
#                     fisher = sp.stats.fisher_exact(table, **fisher_parameters)
#                     return {"statistic": fisher[0], "pvalue": fisher[1], "data": fisher}
#                 else:
#                     return {"statistic": -1, "pvalue": -1, "data": None}

#             # calc
#             calc_diff("fisher_exact", fisher_f)

#         # pearson chi2; only makes sense for categorical data values
#         if chi2:

#             # set parameters
#             chi2_parameters = {}
#             if isinstance(chi2, dict):
#                 chi2_parameters = chi2

#             # define function
#             def chi2_f(df, labels, c0, c1):
#                 table = df.reset_index().groupby(["bin", "values"]).count().unstack().fillna(0)
#                 chi2 = sp.stats.chi2_contingency(table, **chi2_parameters)
#                 return {"statistic": chi2[0], "pvalue": chi2[1], "data": chi2}

#             # calc
#             calc_diff("chi2", chi2_f)

#         # ranksums; should only be used for continuous distributions
#         if ranksums:

#             def ranksums_f(df, labels, c0, c1):
#                 ranksums = sp.stats.ranksums(c0["values"], c1["values"])
#                 return {"statistic": ranksums[0], "pvalue": ranksums[1]}

#             calc_diff("ranksums", ranksums_f)

#         # mannwhtineyu
#         if mannwhitneyu:

#             # set parameters
#             mannwhitneyu_parameters = {"alternative": "two-sided", "use_continuity": False}
#             if isinstance(mannwhitneyu, dict):
#                 mannwhitneyu_parameters = mannwhitneyu

#             def mannwhitneyu_f(df, labels, c0, c1):
#                 mannwhitneyu = sp.stats.mannwhitneyu(c0["values"], c1["values"], **mannwhitneyu_parameters)
#                 return {"statistic": mannwhitneyu[0], "pvalue": mannwhitneyu[1], "data": mannwhitneyu}
            
#             calc_diff("mannwhitneyu", mannwhitneyu_f)

#         # custom
#         if custom_difference_function is not None:

#             if not isinstance(custom_difference_function, list):
#                 custom_difference_function = [("custom", custom_difference_function)]

#             for name, difference_function in custom_difference_function:
#                 calc_diff(name, difference_function)

#     # multiple comparison

#     if isinstance(multipletests, str):
#         multipletests = [multipletests]
#     tests = differences[next(iter(differences))]["tests"].keys()
#     columns = list(differences.keys())
#     for test in tests:
#         pvalues = [differences[c]["tests"][test]["pvalue"] for c in columns]
#         for method in multipletests:
#             _, corrected_pvalues, _, _ = statsmodels.stats.multitest.multipletests(pvalues, method=method)
#             for c, p in zip(columns, corrected_pvalues):
#                 differences[c]["tests"][test][f"pvalue___{method}"] = p

#     # return
#     if return_dataframe:

#         df = pd.DataFrame({("column_name", "", ""): sorted(differences.keys())})

#         # add stats
#         for label in differences[c]["stats"].keys():
#             for stat in differences[c]["stats"][label].keys():
#                 df[("stats", label, stat)] = [differences[c]["stats"][label][stat] for c in df.column_name]

#         # add tests

#         def add_test(name):
#             for key, value in differences[df.column_name[0]]["tests"][name].items():
#                 if key != "___sampled" and key != "data":
#                     df[("tests", name, key)] = [differences[c]["tests"][name][key] for c in df.column_name]

#             if samples_fraction is not None:
#                 for key, value in differences[df.column_name[0]]["tests"][name]["___sampled"].items():
#                     if key != "data":
#                         df[("tests", name + "___sampled", key)] = [
#                             differences[c]["tests"][name]["___sampled"][key] for c in df.column_name]

#         if chi2:
#             add_test("chi2")
#         if ranksums:
#             add_test("ranksums")
#         if mannwhitneyu:
#             add_test("mannwhitneyu")
#         if custom_difference_function is not None:
#             if isinstance(custom_difference_function, list):
#                 for name, _ in custom_difference_function:
#                     add_test(name)
#             else:
#                 add_test("custom")

#         df.columns = pd.MultiIndex.from_tuples(df.columns)
#         df.index = pd.MultiIndex.from_tuples([(c if isinstance(c, tuple) else (c,)) for c in df["column_name"]])
#         df.drop(columns=["column_name"], inplace=True)
#         # set_index("column_name", inplace=True)

#         # return
#         return differences, df

#     else:
#         return differences


# def continuous_difference(
#         target,
#         data,
#         spearman=True,
#         pearson=False,
#         kendall=False,
#         custom_difference_function=None,
#         p1_on_error=False,
#         multipletests="bonferroni",
#         exclude_target=True,
#         return_dataframe=False,
#         samples_fraction=None,
#         samples_n=100,
#         overall_stats_label="__overall__",
#         drop_nans=True):
#     """Given a binary label for each row,
#     this function tests if the values in each each column grouped by these labels
#     come from the same distribution.
#     Various test are supported including Pearson's Chi2, Ranksums (Wilcoxon), and Mann-Whitney-U.

#     Parameters
#     ----------
#     target : str or np.array or pd.Series
#         continuous target
#     data : pd.DataFrame
#         the columns to compare (may include binary labels)
#     spearman : bool
#         whether to calculate Spearman correlation
#     pearson : bool
#         whether to calculate Pearson correlation
#     custom_difference_function : difference function or list of name-difference function tuples ([("custom1", diff_function1), ...]), optional
#         difference_function(
#             pd.DataFrame(values, target_data):
#                 return  { "statistic": chi2[0], "pvalue": chi2[1], "data": chi2 }
#         (the default is None, which results in no additional tests)
#     p1_on_error : bool, optional
#         If set to `True` and an error is raised,
#         we set `{ "statistic": None, "pvalue": 1, "error": True, "data": <exception> }`
#         (the default is False, which raises each exception/warning)
#     multipletests : str or list of str
#         Multiple comparison tests for pvalues of statistical tests
#         (cf. https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html)
#     exclude_target : bool, optional
#         If the labels are part of the data and this is set to `False`,
#         we include the label in the results
#         (the default is True, which excludes the labels)
#     return_dataframe: bool, optional
#         If set to true, the method returns a simplified version of this as a pandas.DataFrame

#     Returns
#     -------
#     dict
#         {
#             "column_name": {
#                 "stats": {
#                     "<value1>": {
#                         "std": 123,
#                         "mean": 123,
#                         "median": 123,
#                         "n": 123,
#                     },
#                     "<value2>": { ... }
#                 },
#                 "tests": {
#                     "test": {
#                         "statistic": 123,
#                         "pvalue": 0.05,
#                         "data": <complete test result>,
#                         ["error": True] # optional; only set if True
#                     }
#                 }
#             }
#         }
#     """

#     if isinstance(target, str):
#         if data is None:
#             raise Exception("If 'binary_labels' is a String, 'data' must be set.")
#         target_data = data[target]
#     else:
#         target_data = target

#     differences = {}

#     for c in data.columns:

#         # skip target column
#         if exclude_target and isinstance(target, str) and c == target:
#             continue

#         # values
#         values = data[c]

#         # init differences
#         differences[c] = {
#             "stats": {
#                 overall_stats_label: collections.OrderedDict([
#                     ("median",      np.nanmedian(values)),
#                     ("mean",        np.nanmean(values)),
#                     ("std",         np.nanstd(values)),
#                     ("n",           (~np.isnan(values)).sum()),
#                     ("n_na",        values.isna().sum()),
#                     ("n_unique",    len(values[~np.isnan(values)].unique())),
#                     ("binary",      len(values[~np.isnan(values)].unique()) == 2),
#                     ("unary",       len(values[~np.isnan(values)].unique()) == 1)])
#             },
#             "tests": collections.OrderedDict()
#         }

#         # tests
#         difference_map = {}
#         differences[c]["tests"] = difference_map

#         def calc_diff_generic(difference_function, values, target_data):

#             select = ~np.isnan(values) & ~np.isnan(target_data)

#             try:
#                 return difference_function(values[select], target_data[select])
#             except Exception as e:
#                 if p1_on_error:
#                     return {"statistic": None, "pvalue": 1, "error": True, "data": e}
#                 else:
#                     raise e

#         def sample_diff(name, difference_function):

#             split = sklearn.model_selection.StratifiedShuffleSplit(
#                 n_splits=samples_n, train_size=samples_fraction, test_size=1 - samples_fraction)
#             samples = [s for s, _ in split.split(values, target_data)]

#             results = [calc_diff_generic(difference_function, values[idx], target_data[idx]) for idx in samples]

#             return {
#                 "statistic": np.mean([s["statistic"] for s in results]),
#                 "statistic_std": np.std([s["statistic"] for s in results]),
#                 "pvalue": np.mean([s["pvalue"] for s in results]),
#                 "pvalue_std": np.std([s["pvalue"] for s in results]),
#                 "data": {
#                     "samples": results,
#                     "splits": samples
#                 }
#             }

#         def calc_diff(name, difference_function):

#             difference_map[name] = calc_diff_generic(difference_function, values, target_data)

#             if samples_fraction is not None:
#                 difference_map[name]["___sampled"] = sample_diff(name, difference_function)

#         # spearman; only works on binary data values
#         if spearman:

#             # define function
#             def spearman_f(values, target_data):
#                 spearman = sp.stats.spearmanr(values, target_data)
#                 return {"statistic": spearman[0], "pvalue": spearman[1], "data": spearman}

#             # calc
#             calc_diff("spearman", spearman_f)

#         # pearson; only works on binary data values
#         if pearson:

#             # define function
#             def pearson_f(values, target_data):
#                 pearson = sp.stats.pearsonr(values, target_data)
#                 return {"statistic": pearson[0], "pvalue": pearson[1], "data": pearson}

#             # calc
#             calc_diff("pearson", pearson_f)

#         # pearson; only works on binary data values
#         if kendall:

#             # define function
#             def kendall_f(values, target_data):
#                 kendall = sp.stats.kendalltau(values, target_data)
#                 return {"statistic": kendall[0], "pvalue": kendall[1], "data": kendall}

#             # calc
#             calc_diff("kendall", kendall_f)

#         # custom
#         if custom_difference_function is not None:

#             if not isinstance(custom_difference_function, list):
#                 custom_difference_function = [("custom", custom_difference_function)]

#             for name, difference_function in custom_difference_function:
#                 calc_diff(name, difference_function)

#     # multiple comparison

#     if isinstance(multipletests, str):
#         multipletests = [multipletests]
#     tests = differences[next(iter(differences))]["tests"].keys()
#     columns = list(differences.keys())
#     for test in tests:
#         pvalues = [differences[c]["tests"][test]["pvalue"] for c in columns]
#         if any([np.isnan(p) for p in pvalues]):
#             warnings.warn(f"NA replaced by 1 for multiple hypothesis comparison correction ({test}).")
#         pvalues = [(1 if np.isnan(p) else p) for p in pvalues]
#         for method in multipletests:
#             _, corrected_pvalues, _, _ = statsmodels.stats.multitest.multipletests(pvalues, method=method)
#             for c, p in zip(columns, corrected_pvalues):
#                 differences[c]["tests"][test][f"pvalue_{method}"] = p

#     # return
#     if return_dataframe:

#         df = pd.DataFrame({("column_name", "", ""): sorted(differences.keys())})

#         # add stats
#         for label in differences[c]["stats"].keys():
#             for stat in differences[c]["stats"][label].keys():
#                 df[("stats", label, stat)] = [differences[c]["stats"][label][stat] for c in df.column_name]

#         # add tests

#         def add_test(name):
#             for key, value in differences[df.column_name[0]]["tests"][name].items():
#                 if key != "___sampled" and key != "data":
#                     df[("tests", name, key)] = [differences[c]["tests"][name][key] for c in df.column_name]

#             if samples_fraction is not None:
#                 for key, value in differences[df.column_name[0]]["tests"][name]["___sampled"].items():
#                     if key != "data":
#                         df[("tests", name + "___sampled", key)] = [
#                             differences[c]["tests"][name]["___sampled"][key] for c in df.column_name]

#         if spearman:
#             add_test("spearman")
#         if pearson:
#             add_test("pearson")
#         if kendall:
#             add_test("kendall")
#         if custom_difference_function is not None:
#             if isinstance(custom_difference_function, list):
#                 for name, _ in custom_difference_function:
#                     add_test(name)
#             else:
#                 add_test("custom")

#         df.columns = pd.MultiIndex.from_tuples(df.columns)
#         df.index = pd.MultiIndex.from_tuples([(c if isinstance(c, tuple) else (c,)) for c in df["column_name"]])
#         df.drop(columns=["column_name"], inplace=True)

#         # return
#         return differences, df

#     else:
#         return differences


def column_info_data_format():
    """Column info data format:
    {
        "registry": {
            "correlations": {

            },
            clusterings: {
                "kmeans": {
                    # clustering info
                }
            },
            "coordinates": {
                "name": {
                    # info
                }
            },
            "targets": {
                "name": {
                    #info
                }
            },
        },
        columns: {
            "column1": {
                "correlations": {
                    "pearson": {
                        "column1": {
                            "correlation": 0.9,
                            "pvalue": 0.05,
                            # anything else
                        },
                        # ...
                    },
                    # ...
                },
                "clusters": {
                    "kmeans": {
                        "id": int,
                        "color": #123abe,
                        "center": np.array([...])
                    },
                    # ...
                }
                "coordinates": {
                    "tsne_abs": np.array([0.5, 0.2]),
                    # ...
                },
                "targets": {
                    "target_name": {  
                        # target data
                    }
                }
            },
            # ... 
        },
        # ...
    }
    """
    pass


def column_infos_add_correlations(
        column_infos, 
        data,
        correlation_name="default",
        correlation_function='spearman',
        homogeneous_data_policy=None,
        nan_policy=False,
        n_jobs=None,
        verbose=False,
        add_to_coordinates_with_name="correlations_default"):

    # calculate correlations
    corr_info = calculate_correlation_parallel(
        data.values,
        correlation_function=correlation_function,
        rowvars=False,
        homogeneous_data_policy=homogeneous_data_policy,
        nan_policy=nan_policy,
        n_jobs=n_jobs,
        verbose=verbose)

    # add correlations to overview
    dicts.recursive_set(
        column_infos, 
        ["registry", "correlations", correlation_name], 
        collections.OrderedDict())

    # add correlations to column info
    column_infos.setdefault("columns", {})
    for src, correlations, pvalues, overlap \
            in zip(data.columns, corr_info.statistics, corr_info.pvalues, corr_info.overlap):

        # set correlations

        dicts.recursive_set(
            column_infos,
            ["columns", src, "correlations", correlation_name],
            collections.OrderedDict([
                (dst, {
                    "statistic": correlations[0, dst_idx],
                    "pvalue": pvalues[0, dst_idx],
                    "overlap": overlap[0, dst_idx],
                }) for dst_idx, dst in enumerate(data.columns)
            ]))

        # set coordinates
        if add_to_coordinates_with_name:
            dicts.recursive_set(
                column_infos,
                ["columns", src, "coordinates", add_to_coordinates_with_name],
                correlations)
            
    return column_infos


def column_infos_add_clustering(
        column_infos, 
        clustering_coordinates="correlations_default",
        clustering_algorithm='kmeans',
        clustering_name="default", 
        clustering_parameters=None,
        color_palette=None,
        cluster_on_absolute_correlation_values=True,
        nan_policy=None):

    # set default behavior
    if clustering_algorithm == "kmeans":
        clustering_algorithm = sklearn.cluster.KMeans
        if clustering_parameters is None:
            clustering_parameters = {"n_clusters": 10}

    # extract coordinates for clustering
    columns = sorted(column_infos["columns"].keys())
    coordinates = scipy.sparse.vstack([
        column_infos["columns"][c]["coordinates"][clustering_coordinates] for c in columns]).toarray()

    if nan_policy is not None and np.isnan(coordinates).sum() > 0:
        if nan_policy == "median":
            coordinates = sklearn.impute.SimpleImputer(strategy="median").fit_transform(coordinates)
        elif nan_policy == "mean":
            coordinates = sklearn.impute.SimpleImputer(strategy="mean").fit_transform(coordinates)
        else:
            raise ValueError(f"Unknown NaN handling strategy: {nan_policy}")

    # cluster
    clustering_instance = clustering_algorithm(**clustering_parameters)
    if cluster_on_absolute_correlation_values:
        clustering_instance.fit(abs(coordinates))
    else:
        clustering_instance.fit(coordinates)

    # derive colormap
    
    cluster_labels = sorted(np.unique(clustering_instance.labels_))

    if color_palette is None:
        color_palette = sns.color_palette("hls", len(cluster_labels)).as_hex()
    if isinstance(color_palette, str):
        color_palette = sns.color_palette(color_palette, len(cluster_labels)).as_hex()
    
    color_map = {l: c for l, c in zip(cluster_labels, color_palette)}

    # center map
    if "cluster_centers_" in dir(clustering_instance):
        center_map = {l: coords for l, coords in enumerate(clustering_instance.cluster_centers_)}
    else:
        center_map = None
    
    # add cluster info
    for l in cluster_labels:
        dicts.recursive_set(
            column_infos,
            ["registry", "clusterings", clustering_name, l],
            {
                "color": color_map[l],
                "center": center_map[l] if center_map is not None else None
            })

    # set cluster ids on columns
    for c, label in zip(columns, clustering_instance.labels_):
        dicts.recursive_set(
            column_infos,
            [ "columns", c, "clusters", clustering_name ],
            { 
                "cluster_id": label,
                "cluster_color": color_map[label] 
            })

    return column_infos


def column_infos_add_embedding(
        column_infos,
        embedding_coordinates="correlations_default", 
        embedding_algorithm=None, 
        embedding_name="embedding_default",
        embedding_parameters=None, 
        embedd_on_absolute_correlation_values=True,
        nan_policy=None):

    # set default behavior
    if embedding_algorithm is None:
        embedding_algorithm = sklearn.manifold.TSNE
        if embedding_parameters is None:
            embedding_parameters = {"perplexity": 16}
    elif embedding_algorithm == "pca":
        embedding_algorithm = sklearn.decomposition.PCA
    elif embedding_algorithm == "kpca":
        embedding_algorithm = sklearn.decomposition.KernelPCA

    if embedding_parameters is None:
        embedding_parameters = {}
    
    # extract coordinates for clustering
    columns = sorted(column_infos["columns"].keys())
    coordinates = scipy.sparse.vstack([
        column_infos["columns"][c]["coordinates"][embedding_coordinates] for c in columns]).toarray()

    if nan_policy is not None and np.isnan(coordinates).sum() > 0:
        if nan_policy == "median":
            coordinates = sklearn.impute.SimpleImputer(strategy="median").fit_transform(coordinates)
        elif nan_policy == "mean":
            coordinates = sklearn.impute.SimpleImputer(strategy="mean").fit_transform(coordinates)
        else:
            raise ValueError(f"Unknown NaN handling strategy: {nan_policy}")

    # preprocess 
    if embedd_on_absolute_correlation_values:
        coordinates = abs(coordinates)

    # embedd
    embeddings = embedding_algorithm(**embedding_parameters).fit_transform(coordinates)

    # normalize embedding
    embeddings /= abs(embeddings).max()

    # add embeddings to overview
    dicts.recursive_set(
        column_infos,
        ["registry", "coordinates", embedding_name],
        embedding_parameters)

    # incorporate embeddings into column info
    for c, embedding in zip(columns, embeddings):
        dicts.recursive_set(
            column_infos,
            ["columns", c, "coordinates", embedding_name],
            embedding)
    
    return column_infos


def column_infos_add_target_difference(
        column_infos, 
        data, 
        target, 
        target_name="default",
        **kwargs):

    # calculate target values
    target_data = data[target] if isinstance(target, str) else target
    target_values = calculate_pairwise_comparisons(data, target_data, **kwargs)

    # add target to overview
    dicts.recursive_set(
        column_infos,
        ["registry", "targets", target_name],
        {})

    # add target info to individual columns
    for c, values in target_values.items():
        dicts.recursive_set(
            column_infos,    
            ["columns", c, "targets", target_name],
            values)

    return column_infos


def column_infos_prepare(
        data, 

        column_infos=None,

        correlation_parameters=None,
        clustering_parameters=None,
        embedding_parameters=None,
        target=None,
        target_parameters=None,

        overwrite_correlations=False,
        overwrite_clustering=False,
        overwrite_embedding=False,
        overwrite_target=False,

        random_state=None,
        verbose=False):

    # set default parameters
    # TODO
    if correlation_parameters is None:
        correlation_parameters = {}
    if clustering_parameters is None:
        clustering_parameters = {}
    if embedding_parameters is None:
        embedding_parameters = {}
    if target_parameters is None:
        target_parameters = {}

    if column_infos is None:
        column_infos = collections.OrderedDict()

    def print_verbose(message):
        if verbose:
            print(message)

    if overwrite_correlations or dicts.recursive_get(column_infos, ["registry", "correlations", "default"]) is None:
        print_verbose("* Adding correlations")
        column_infos_add_correlations(column_infos, data, **correlation_parameters)
        # print_verbose(column_infos)
    
    if overwrite_clustering or dicts.recursive_get(column_infos, ["registry", "clusterings", "default"]) is None:
        print_verbose("* Adding clustering")
        column_infos_add_clustering(column_infos, **clustering_parameters)
        # print_verbose(column_infos)
    
    if overwrite_embedding or dicts.recursive_get(column_infos, ["registry", "coordinates", "embedding_default"]) is None:
        print_verbose("* Adding embedding")
        column_infos_add_embedding(column_infos, **embedding_parameters)
        # print_verbose(column_infos)
    
    if target is not None:
        if overwrite_target or dicts.recursive_get(column_infos, ["registry", "targets", "default"]) is None:
            print_verbose("* Adding target")
            column_infos_add_target_difference(column_infos, data, target, **target_parameters)

            # print_verbose(column_infos)

    return column_infos


def column_infos_plot(
        column_infos,
        correlation_name="default",
        edge_include_when_abs_correlation_greater_equals=0.3,
        target_name="default",
        target_component="ranksums",
        cluster_name="default",
        coordinates_name="embedding_default",
        init_matplotlib=True):
    """This is basically a template to copy, paste and customize."""

    # init nodes
    nodes = column_infos["columns"].items()

    # init edges
    edges = [ 
        (src, dst, { 
                "correlation": dst_d["statistic"], 
                "pvalue": dst_d["pvalue"] }) 
            for src, src_d in column_infos["columns"].items()
            for dst, dst_d in src_d["correlations"][correlation_name].items() 
        if abs(dst_d["statistic"]) >= edge_include_when_abs_correlation_greater_equals
    ]

    # node color by cluster
    def node_color(g, n, d):

        if "targets" in d and target_name in d["targets"]:

            if d["targets"][target_name]["tests"][target_component]["pvalue"] >= 0.05:
                return "lightgrey"
            elif "clusters" in d and cluster_name in d["clusters"]:
                return d["clusters"][cluster_name]["cluster_color"]

        # basically an error due to missing values
        return "black"

    # node shape by statistic
    def node_shape(g,n,d):
        if "targets" in d and target_name in d["targets"]:
            if d["targets"][target_name]["tests"][target_component]["statistic"] > 0:
                return "o"
            else:
                return "<"
        else:
            return "d"

    # node size by pvalue of target
    def node_size(g,n,d):
        if "targets" in d and target_name in d["targets"]:
            p = d["targets"][target_name]["tests"][target_component]["pvalue"]
            return 1000 * -np.log10(p)
        else:
            return 200

    # edge color by correlations between variables
    def edge_color(g,src,dst,d):

        correlation = d["correlation"]

        if  correlation > 0:
            if correlation >= 0.5:
                return "red"
            else:
                return "orange"
        elif correlation < 0:
            if correlation <= -0.5:
                return "blue"
            else:
                return "purple"
        else:
            return "black"

    if init_matplotlib:
        plt.figure(figsize=(16,16))
        plt.axis('off')
        plt.xlim([-1.1,1.1])
        plt.ylim([-1.1,1.1])

    nx_plot(
        nodes=nodes,
        nodes_pos=lambda g,n,d: d["coordinates"][coordinates_name],
        nodes_args={
            "node_color": node_color,
            "node_size": node_size,
            "node_shape": node_shape
        },
        edges=edges,
        edges_args={
            "edge_color": edge_color,
            "width": 1
        },
        nodes_labels=lambda g,n,d: n,
        nodes_labels_pos=lambda g,n,d: d["coordinates"][coordinates_name],
        nodes_labels_args={
            "horizontalalignment": "left"
        }
    )

    return column_infos


def column_infos_extract_df(column_infos, columns=None, column_names=None, values=None, target_name="default"):

    if columns is None:
        columns = sorted(column_infos["columns"].keys()) 

    # set default values if none are given
    # if values is None:
    #
    #     some_column = column_infos["columns"][columns[0]]
    #     default_target = some_column["targets"][target_name]
    #
    #     values = [(k, ["targets", target_name] + k)
    #               for k, v in dicts.recursive_items(default_target) if k[-1] != "data"]

    if values is None:
        values = []
        for c in columns:
            for k, v in dicts.recursive_items(column_infos["columns"][c]["targets"][target_name]):
                if k[-1] not in ["data"]:
                    value = (k, ["targets", target_name] + k)
                    if value not in values:
                        values.append(value)

    # build column index for DataFrame
    max_depth = max([len(k) for k, _ in values])
    def format_key(key):
        if isinstance(key, str):
            return key
        elif isinstance(key, list):
            if len(key) is max_depth:
                return key
            else:
                return key + [""] * (max_depth - len(key))

    if max_depth is 1:
        column_index = [k for k, _ in values]
    else:
        column_index = pd.MultiIndex.from_tuples([format_key(k) for k, _ in values])

    # helper function for extractor
    def extract(key, extractor):
        
        if isinstance(extractor, str):
            extractor = [extractor]

        if isinstance(extractor, list):
            return [dicts.recursive_get(column_infos["columns"][c], extractor) for c in columns]

        elif callable(extractor):
            return [extractor(column_infos, column_infos["columns"][c], key) for c in columns]

    # create DataFrame
    df = pd.DataFrame({
        idx: extract(key, extractor) 
        for idx, (key, extractor) in enumerate(values)
    })

    # set row names
    if column_names is None:
        df.index = columns
    elif isinstance(column_names, list):
        df.index = column_names
    elif isinstance(column_names, dict):
        df.index = [column_names[c] for c in columns]

    # set column index
    df.columns = column_index

    return df


def coef_extract_stats(experiment):
    
    # get weights per split
    
    weights = None
    indices = None
    for i_split, split in enumerate(experiment["splits"]):
        
        # get coefficients for split
        m = split["train"]["model"]["object"]
        coef = m.coef_.squeeze()
        
        # init weights
        if weights is None:
            weights = [[] for _ in range(len(coef))]
            indices = [[] for _ in range(len(coef))]
            
        # add weights
        for coef_i, w in enumerate(coef):
            if abs(w) > 0:
                indices[coef_i].append(i_split)
                weights[coef_i].append(w)
        
    weights = [ np.array(ws) for ws in weights ]
    indices = [ np.array(xs) for xs in indices ]
    stats = [ {
        "count": len(ws),
        "count_gt_0": np.sum(ws > 0),
        "count_lt_0": np.sum(ws < 0),

        "mean": np.mean(ws) if len(ws) > 0 else None,
        "median": np.median(ws) if len(ws) > 0 else None,
        "std": np.std(ws) if len(ws) > 0 else None,
        "min": np.min(ws) if len(ws) > 0 else None,
        "max": np.max(ws) if len(ws) > 0 else None,

        "abs_mean": np.mean(abs(ws)) if len(ws) > 0 else None,
        "abs_median": np.median(abs(ws)) if len(ws) > 0 else None,
        "abs_std": np.std(abs(ws)) if len(ws) > 0 else None,
        "abs_min": np.min(abs(ws)) if len(ws) > 0 else None,
        "abs_max": np.max(abs(ws)) if len(ws) > 0 else None,
    } for ws in weights ]
    
    return stats, weights, indices


def coef_plot_stats(experiment, column_order, top=10, sort_by="count", high_to_low=True, models=None):
    """
    We can also plot directly from weights, e.g., using: 
    ```
    plt.plot(sorted(list([ d["count"] for d in  weights.values() if d["count"] > 0])))
    ```
    """

    import numpy as np
    import matplotlib
    import matplotlib.pyplot as plt
    
    stats, weights, indices = coef_extract_stats(experiment)
    weight_map = { c:s for c,s in zip(column_order, stats) } 

    # filter weight map
    weight_map = { k:v for k,v in weight_map.items() if v["count"] > 0 }
    
    # set sort function
    sort_function = lambda a: weight_map[a]["count"]
    if type(sort_by) is str:
        sort_function = lambda a: weight_map[a][sort_by]
    if callable(sort_by):
        sort_function = sort_by
    
    # sort
    weight_names = sorted(
        [ k for k,_ in weight_map.items() ],
        key=sort_function)
    if high_to_low:
        weight_names = list(reversed(weight_names))
    
    # select top weights
    if top is not None:
        weight_names = weight_names[0:top]
    
    fig, ax = plt.subplots(1,2)
    
    # plot average weights
    ax[0].set_title("average weights of non-zero features")
    ax[0].set_ylabel("weight")
    ax[0].set_xlabel("feature")
    im = ax[0].bar(weight_names, [ weight_map[f]["mean"] for f in weight_names ])
    ax[0].set_xticklabels(weight_names, rotation=270)
    
    # plot how often each feature was used (/was non-zero)
    ax[1].set_title("counts of non-zero features")
    ax[1].set_ylabel("count")
    ax[1].set_xlabel("feature")
    im = ax[1].bar(weight_names, [ weight_map[f]["count"] for f in weight_names ])
    ax[1].set_xticklabels(weight_names, rotation=270)


def column_infos_add_target_feature_importance(
        column_infos, column_order, experiment,
        target_name="feature_importance", integrate_weights=False, integrate_indices=False):
    
    stats, weights, indices = coef_extract_stats(experiment)
    for c, s, w, i in zip(column_order, stats, weights, indices):
        dicts.recursive_set(
            column_infos,
            ["columns", c, "targets", target_name],
            {
                "stats": s,
                "weights": w if integrate_weights else None,
                "indices": i if integrate_indices else None
            })

    return column_infos


def visualize_dimensions(
        data, 
        limits=None, 
        figsize=(7, 7),
        dim_labels=None,
        max_dims=None,

        plots=None,
        force_limits=False,

        bins=20, 
        hist_norm=True,
        scatter_labels=None,
        scatter_origin=False,
        hist_origin=False,
        
        hist_params=None,
        scatter_params=None,
        hexbin_params=None):

    # derive number of dimensions to plot
    n_dims = data.shape[1]
    if max_dims is not None:
        n_dims = min(n_dims, max_dims+1)

    # init plots
    if plots is None:
        plots = ["scatter"]
    elif isinstance(plots, str):
        plots = [plots]

    # number of plot types
    n_plot_types = len(plots)
    grid = gridspec.GridSpec(1, n_plot_types)

    def plot_scatter(ax, data, row, col):
        params = {
            "s": 0.5,
            "c": scatter_labels
        }
        if scatter_params is not None:
            params = { **params, **scatter_params }

        if scatter_params is not None and "marker" in scatter_params and not isinstance(scatter_params["marker"], str):
            
            markers = scatter_params["marker"]
            unique_markers = np.unique(markers)
            del params["marker"]

            for marker in unique_markers:

                if "c" in params and not isinstance(params["c"], str):
                    c = params["c"]
                    del params["c"]
                    idx = [ i for i,m in enumerate(markers) if m == marker ]
                    ax.scatter(data[idx,col], data[idx,row], marker=marker, c=c[idx], **params)
                    params["c"] = c
                else:
                    idx = [ i for i,m in enumerate(markers) if m == marker ]
                    ax.scatter(data[idx,col], data[idx,row], marker=marker, **params)
        
        else:
            ax.scatter(data[:,col], data[:,row], **params)

        if scatter_origin:
            plt.axhline(0, ls="--", c="black", linewidth=0.5)
            plt.axvline(0, ls="--", c="black", linewidth=0.5)

    def plot_hexbin(ax, data, row, col):
        params = {
            "gridsize": bins
        }
        if hexbin_params is not None:
            params = { **params, **hexbin_params }
            
        ax.hexbin(data[:, col], data[:, row], **params)

    def plot_hist(ax, data, row, col):
        params = {
            "bins": bins,
            "norm": matplotlib.colors.LogNorm() if hist_norm else None
        }
        if hist_params is not None:
            params = { **params, **hist_params }

        ax.hist2d(data[:,col], data[:,row], **params)

        if hist_origin:
            plt.axhline(0, ls="--", c="black", linewidth=0.5)
            plt.axvline(0, ls="--", c="black", linewidth=0.5)

    # init plot info
    plots_info = []
    for i, plot in enumerate(plots):

        name = str(i)
        plot_spec = gridspec.GridSpecFromSubplotSpec(n_dims-1, n_dims-1, subplot_spec=grid[i])
        plot_func = plot

        if isinstance(plot, tuple):
            name = plot[0]
            plot_func = plot[1]
        
        elif isinstance(plot, str):
            if plot == "scatter":
                name = "scatter"
                plot_func = plot_scatter
            elif plot == "hexbin":
                name = "hexbin"
                plot_func = plot_hexbin
            elif plot == "hist":
                name = "hist"
                plot_func = plot_hist
            else:
                raise Exception("Unknown plot type: {}".format(plot))

        plots_info.append((name, plot_spec, plot_func))

    # init axes container
    axes = { name: np.empty((n_dims, n_dims), dtype="object") for name, _, _ in plots_info }
    
    fig = plt.figure(figsize=figsize)   
    # plt.rcParams['axes.labelweight'] = 'bold' 

    for row in range(0, n_dims-1):
        for col in range(1, n_dims):
            
            d_row = row if dim_labels is None else dim_labels[row]
            d_col = col if dim_labels is None else dim_labels[col]

            # TODO: results in cutoffs
            if limits is None:
                limits = [[min(data[:, col]), max(data[:, col])], [min(data[:, row]), max(data[:, row])]]

            for plot_name, plot_spec, plot_func in plots_info:

                ax = fig.add_subplot(plot_spec[row, col-1])
                axes[plot_name][row,col] = ax
                if col <= row:
                    ax.axis("off")
                    continue
        
                if row is 0:
                    ax.xaxis.set_label_position('top') 
                    ax.xaxis.tick_top()
                    ax.set_xlabel(d_col)
                else:
                    ax.get_xaxis().set_ticklabels([])

                if row == col-1:
                    ax.set_ylabel(d_row)
                else:
                    ax.get_yaxis().set_ticklabels([])
                
                plot_func(ax, data, row, col)

                if force_limits:
                    ax.set_xlim(limits[0])
                    ax.set_ylim(limits[1])

    return fig, axes
            

def embedd_and_plot(
        data,
        y=None,
        algorithm="pca", algorithm_params=None,
        dim_x=0, dim_y=1,
        annotations=None,
        annotation_params=None,
        palette=None,
        plot_params=None,
        subsample=None):
    """
    Embedds a list of datasets and scatter plots them.

    TODO:
    - [ ] if the data is given as a dict, use keys as labels
    - [ ] return embedded data
    - [ ] support down sampling (fixed number or min of all datasets)
    - [ ] support/use ~.visualize_dimensions

    Parameters
    ----------
    data: list[np.ndarray] or dict of
    y
    algorithm
    algorithm_params
    dim_x
    dim_y
    annotations
    annotation_params
    palette
    plot_params
    subsample

    Returns
    -------

    """

    if algorithm_params is None:
        algorithm_params = {}

    if algorithm == "pca":
        algorithm = sklearn.decomposition.PCA(**algorithm_params)
    elif algorithm == "lda":
        algorithm = sklearn.discriminant_analysis.LinearDiscriminantAnalysis(**algorithm_params)
    elif algorithm == "tsne":
        algorithm = sklearn.manifold.TSNE(**algorithm_params)

    if plot_params is None:
        plot_params = {}

    # derive data set labels
    labels = np.arange(len(data))
    if isinstance(data, dict):
        labels = sorted(data.keys())
        data = [data[label] for label in labels]

    if subsample is not None:
        indices = [np.random.choice(np.arange(d.shape[0]), subsample, replace=False) for d in data]
        data = [d[idx, :] for d, idx in zip(data, indices)]
        if y is not None:
            y = [yy[idx, :] for yy, idx in zip(y, indices)]

    if isinstance(y, list):
        y = np.concatenate(y)
    embedded = algorithm.fit_transform(np.concatenate(data), y)

    if palette is None:
        palette = sns.color_palette()

    colors = np.repeat(palette[:len(data)], [d.shape[0] for d in data], axis=0)

    sns_params = {
        "palette": palette[:len(data)],
        "s": 2,
        "alpha": 1,
        "linewidth": 0
    }

    sns_params["palette"] = palette
    scatter = sns.scatterplot(
        x=embedded[:, dim_x],
        y=embedded[:, dim_y],
        hue=np.repeat(labels, [d.shape[0] for d in data]),
        **{**sns_params, **plot_params}
    )

    if annotations is not None:
        annotations = np.concatenate(annotations)
        default_annotation_params = {}
        if annotation_params is None:
            annotation_params = {}
        for a, vec in zip(annotations, embedded):
            plt.annotate(a, xy=(vec[dim_x], vec[dim_y]), **{**default_annotation_params, **annotation_params})

    return scatter, embedded


def plot_image(path, ax=None, crop_whitespace=True):
    """
    Plots an image from a file. Can crop whitespace.

    Parameters
    ----------
    path: str
        Path to image file
    ax: matplotlib.axes.Axes, optional, default: None
        Axes to plot on
    crop_whitespace: bool, optional, default: True
        Whether to crop white space around image

    Returns
    -------
    matplotlib.axes.Axes
        The axes the image was plotted on

    """
    import skimage

    img = skimage.img_as_float(matplotlib.image.imread(path))

    if crop_whitespace:
        # crop whitespace: https://stackoverflow.com/questions/26310873/how-do-i-crop-an-image-on-a-white-background-with-python

        # Select all pixels almost equal to white
        # (almost, because there are some edge effects in jpegs
        # so the boundaries may not be exactly white)
        white = np.repeat(1, img.shape[2])
        mask = np.abs(img - white).sum(axis=2) < 0.05

        # Find the bounding box of those pixels
        coords = np.array(np.nonzero(~mask))
        top_left = np.min(coords, axis=1)
        bottom_right = np.max(coords, axis=1)

        img = img[
            top_left[0]:bottom_right[0],
            top_left[1]:bottom_right[1]]

    if ax is None:
        ax = plt.gca()

    ax.imshow(img)
    ax.axis("off")
    return ax


def compare_vectors(vectors, labels=None, plot=True, plot_params=None):

    if labels is None:
        labels = np.arange(len(vectors))

    if plot_params is None:
        plot_params = {"kind": "bar", "row": "label"}

    df = pd.DataFrame({
        "dimension": np.arange(len(vectors[0])),
    })
    for label, v in zip(labels, vectors):
        df[label] = v

    if plot:
        # sns.catplot(x="dimension", y="value", data=df, **plot_params)
        df.set_index("dimension").plot.bar(**plot_params)

    return df

